@@ -1,7 +1,9 @@
-from typing import Callable, Any, Dict, Union, Sequence
-from .apply_base import ApplyDocumentScoringTransformer, ApplyQueryTransformer, ApplyDocFeatureTransformer, ApplyForEachQuery, ApplyGenericTransformer, Transformer
-from nptyping import NDArray
+from functools import partial
+from typing import Callable, Any, Dict, Union, Iterator, Sequence
+import numpy.typing as npt
 import pandas as pd
+import pyterrier as pt
+from pyterrier.apply_base import ApplyDocumentScoringTransformer, ApplyQueryTransformer, ApplyDocFeatureTransformer, ApplyForEachQuery, ApplyGenericTransformer, ApplyIndexer
 
 def _bind(instance, func, as_name=None):
     """
@@ -15,7 +17,7 @@ def _bind(instance, func, as_name=None):
     setattr(instance, as_name, bound_method)
     return bound_method
 
-def query(fn : Callable[[pd.Series], str], *args, **kwargs) -> Transformer:
+def query(fn : Callable[[pd.Series], str], *args, **kwargs) -> pt.Transformer:
     """
         Create a transformer that takes as input a query, and applies a supplied function to compute a new query formulation.
 
@@ -41,12 +43,12 @@ def query(fn : Callable[[pd.Series], str], *args, **kwargs) -> Transformer:
                 return " ".join(terms)
 
             # a query rewriting transformer that applies the _remove_stops to each row of an input dataframe
-            p1 = pt.apply.query(_remove_stops) >> pt.BatchRetrieve(index, wmodel="DPH")
+            p1 = pt.apply.query(_remove_stops) >> pt.terrier.Retriever(index, wmodel="DPH")
 
             # an equivalent query rewriting transformer using an anonymous lambda function
             p2 = pt.apply.query(
                     lambda q :  " ".join([t for t in q["query"].split(" ") if t not in stops ])
-                ) >> pt.BatchRetrieve(index, wmodel="DPH")
+                ) >> pt.terrier.Retriever(index, wmodel="DPH")
 
         In both of the example pipelines above (`p1` and `p2`), the exact topics are not known until the pipeline is invoked, e.g.
         by using `p1.transform(topics)` on a topics dataframe, or within a `pt.Experiment()`. When the pipeline 
@@ -57,7 +59,7 @@ def query(fn : Callable[[pd.Series], str], *args, **kwargs) -> Transformer:
     """
     return ApplyQueryTransformer(fn, *args, **kwargs)
 
-def doc_score(fn : Union[Callable[[pd.Series], float], Callable[[pd.DataFrame], Sequence[float]]], *args, batch_size=None, **kwargs) -> Transformer:
+def doc_score(fn : Union[Callable[[pd.Series], float], Callable[[pd.DataFrame], Sequence[float]]], *args, batch_size=None, **kwargs) -> pt.Transformer:
     """
         Create a transformer that takes as input a ranked documents dataframe, and applies a supplied function to compute a new score.
         Ranks are automatically computed. doc_score() can operate row-wise, or batch-wise, depending on whether batch_size is set.
@@ -73,7 +75,7 @@ def doc_score(fn : Union[Callable[[pd.Series], float], Callable[[pd.DataFrame],
         Example (Row-wise)::
 
             # this transformer will subtract 5 from the score of each document
-            p = pt.BatchRetrieve(index, wmodel="DPH") >> 
+            p = pt.terrier.Retriever(index, wmodel="DPH") >> 
                 pt.apply.doc_score(lambda doc : doc["score"] -5)
 
         Can be used in batch-wise manner, which is particularly useful for appling neural models. In this case,
@@ -83,12 +85,16 @@ def doc_score(fn : Union[Callable[[pd.Series], float], Callable[[pd.DataFrame],
                 # returns series of lengths
                 return df.text.str.len()
             
-            pipe = pt.BatchRetrieve(index) >> pt.apply.doc_score(_doclen, batch_size=128)
+            pipe = pt.terrier.Retriever(index) >> pt.apply.doc_score(_doclen, batch_size=128)
+
+        Can also be used to create individual features that are combined using the ``**`` feature-union operator::
+
+            pipeline = bm25 >> ( some_features ** pt.apply.doc_score(_doclen) )
 
     """
     return ApplyDocumentScoringTransformer(fn, *args, batch_size=batch_size, **kwargs)
 
-def doc_features(fn : Callable[[pd.Series], NDArray[Any]], *args, **kwargs) -> Transformer:
+def doc_features(fn : Callable[[pd.Series], npt.NDArray[Any]], *args, **kwargs) -> pt.Transformer:
     """
         Create a transformer that takes as input a ranked documents dataframe, and applies the supplied function to each document to compute feature scores. 
 
@@ -111,26 +117,54 @@ def doc_features(fn : Callable[[pd.Series], NDArray[Any]], *args, **kwargs) -> T
                 f2 = len(content.split(" "))
                 return np.array([f1, f2])
 
-            p = pt.BatchRetrieve(index, wmodel="BM25") >> 
+            p = pt.terrier.Retriever(index, wmodel="BM25") >> 
                 pt.apply.doc_features(_features )
 
+        NB: If you only want to calculate a single feature to add to existing features, it is better to use ``pt.apply.doc_score()`` 
+        and the ``**`` feature union operator::
+
+            pipeline = bm25 >> ( some_features ** pt.apply.doc_score(one_feature) )
+
     """
     return ApplyDocFeatureTransformer(fn, *args, **kwargs)
 
-def rename(columns : Dict[str,str], *args, **kwargs) -> Transformer:
+def indexer(fn : Callable[[Iterator[Dict[str,Any]]], Any], **kwargs) -> pt.Indexer:
+    """
+        Create an instance of pt.Indexer using a function that takes as input an interable dictionary.
+
+        The supplied function is called once. It may optionally return something (typically a reference to the "index").
+
+        Arguments:
+            fn(Callable): the function that consumed documents.
+
+        Example::
+
+            # make a pt.Indexer that returns the numnber of documents consumed
+            def _counter(iter_dict):
+                count = 0
+                for d in iter_dict:
+                    count += 1
+                return count
+            indexer = pt.apply.indexer(_counter)
+            rtr = indexer.index([ {'docno' : 'd1'}, {'docno' : 'd2'}])
+    """
+    return ApplyIndexer(fn, **kwargs)
+
+def rename(columns : Dict[str,str], *args, errors='raise', **kwargs) -> pt.Transformer:
     """
         Creates a transformer that renames columns in a dataframe. 
 
         Args:
-            columns(dict): A dictionary mapping from old column name to new column name 
+            columns(dict): A dictionary mapping from old column name to new column name
+            errors(str): Maps to df.rename() errors kwarg - default to 'raise', alternatively can be 'ignore'
 
         Example::
             
-            pipe = pt.BatchRetrieve(index, metadata=["docno", "body"]) >> pt.apply.rename({'body':'text'})
+            pipe = pt.terrier.Retriever(index, metadata=["docno", "body"]) >> pt.apply.rename({'body':'text'})
     """
-    return ApplyGenericTransformer(lambda df: df.rename(columns=columns), *args, **kwargs)
+    return ApplyGenericTransformer(lambda df: df.rename(columns=columns, errors=errors), *args, **kwargs)
 
-def generic(fn : Callable[[pd.DataFrame], pd.DataFrame], *args, batch_size=None, **kwargs) -> Transformer:
+def generic(fn : Callable[[pd.DataFrame], pd.DataFrame], *args, batch_size=None, **kwargs) -> pt.Transformer:
     """
         Create a transformer that changes the input dataframe to another dataframe in an unspecified way.
 
@@ -148,12 +182,12 @@ def generic(fn : Callable[[pd.DataFrame], pd.DataFrame], *args, batch_size=None,
             # this transformer will remove all documents at rank greater than 2.
 
             # this pipeline would remove all but the first two documents from a result set
-            pipe = pt.BatchRetrieve(index) >> pt.apply.generic(lambda res : res[res["rank"] < 2])
+            pipe = pt.terrier.Retriever(index) >> pt.apply.generic(lambda res : res[res["rank"] < 2])
 
     """
     return ApplyGenericTransformer(fn, *args, batch_size=batch_size, **kwargs)
 
-def by_query(fn : Callable[[pd.DataFrame], pd.DataFrame], *args, batch_size=None, **kwargs) -> Transformer:
+def by_query(fn : Callable[[pd.DataFrame], pd.DataFrame], *args, batch_size=None, **kwargs) -> pt.Transformer:
     """
         As `pt.apply.generic()` except that fn receives a dataframe for one query at at time, rather than all results at once.
         If batch_size is set, fn will receive no more than batch_size documents for any query. The verbose kwargs controls whether
@@ -167,15 +201,15 @@ class _apply:
         _bind(self, lambda self, fn, *args, **kwargs : query(fn, *args, **kwargs), as_name='query')
         _bind(self, lambda self, fn, *args, **kwargs : doc_score(fn, *args, **kwargs), as_name='doc_score')
         _bind(self, lambda self, fn, *args, **kwargs : doc_features(fn, *args, **kwargs), as_name='doc_features')
+        _bind(self, lambda self, fn, *args, **kwargs : indexer(fn, *args, **kwargs), as_name='indexer')
         _bind(self, lambda self, fn, *args, **kwargs : rename(fn, *args, **kwargs), as_name='rename')
         _bind(self, lambda self, fn, *args, **kwargs : by_query(fn, *args, **kwargs), as_name='by_query')
-        _bind(self, lambda self, fn, *args, **kwargs : generic(fn, *args, **kwargs), as_name='generic')
+        _bind(self, lambda self, fn, *args, **kwargs : generic(fn, *args, **kwargs), as_name='generic')     
     
     def __getattr__(self, item):
-        from functools import partial
         return partial(generic_apply, item)
 
-def generic_apply(name, *args, drop=False, **kwargs) -> Transformer:
+def generic_apply(name, *args, drop=False, **kwargs) -> pt.Transformer:
     if drop:
         return ApplyGenericTransformer(lambda df : df.drop(name, axis=1), *args, **kwargs) 
     
