@@ -1,7 +1,8 @@
-from .transformer import Transformer
-from . import tqdm
+
+from .transformer import Transformer, Indexer
 from .model import add_ranks, split_df
 import pandas as pd
+import pyterrier as pt
 
 class ApplyTransformerBase(Transformer):
     """
@@ -34,14 +35,17 @@ class ApplyForEachQuery(ApplyTransformerBase):
 
         import math, pandas as pd
         from pyterrier.model import split_df
-        from . import tqdm
 
         it = res.groupby("qid")
+        lastqid = None
         if self.verbose:
-            it = tqdm(it, unit='query')
+            it = pt.tqdm(it, unit='query')
         try:
             if self.batch_size is None:
-                query_dfs = [self.fn(group) for qid, group in it]
+                query_dfs = []
+                for qid, group in it:
+                    lastqid = qid
+                    query_dfs.append(self.fn(group))
             else:
                 # fn cannot be applied to more than batch_size rows at once
                 # so we must split and reconstruct the output FOR EACH QUERY
@@ -52,7 +56,7 @@ class ApplyForEachQuery(ApplyTransformerBase):
                     iterator = split_df(group, num_chunks)
                     query_dfs.append( pd.concat([self.fn(chunk_df) for chunk_df in iterator]) )
         except Exception as a:
-            raise Exception("Problem applying %s" % self.fn) from a
+            raise Exception("Problem applying %s for qid %s" % (self.fn, lastqid)) from a
 
         if self.add_ranks:
             try:
@@ -60,8 +64,8 @@ class ApplyForEachQuery(ApplyTransformerBase):
             except KeyError as ke:
                 suffix = 'Try setting add_ranks=False'
                 if len(query_dfs) > 0 and 'score' not in query_dfs[0].columns:
-                    suffix = 'Score column not present. Set add_ranks=False'
-                raise ValueError("Cannot apply add_ranks in pt.apply.by_query -" + suffix) from ke
+                    suffix = 'score column not present. Set add_ranks=False'
+                raise ValueError("Cannot apply add_ranks in pt.apply.by_query - " + suffix) from ke
         rtr = pd.concat(query_dfs)
         return rtr
 
@@ -75,7 +79,7 @@ class ApplyDocumentScoringTransformer(ApplyTransformerBase):
             def _score_fn(row):
                 return float(row["url".count("/")])
             
-            pipe = pt.BatchRetrieve(index) >> pt.apply.doc_score(_score_fn)
+            pipe = pt.terrier.Retriever(index) >> pt.apply.doc_score(_score_fn)
 
         Can be used in batching manner, which is particularly useful for appling neural models. In this case,
         the scoring function receives a dataframe, rather than a single row::
@@ -83,7 +87,7 @@ class ApplyDocumentScoringTransformer(ApplyTransformerBase):
             def _doclen(df):
                 return df.text.str.len()
             
-            pipe = pt.BatchRetrieve(index) >> pt.apply.doc_score(_doclen)
+            pipe = pt.terrier.Retriever(index) >> pt.apply.doc_score(_doclen)
 
     """
     def __init__(self, fn,  *args, batch_size=None, **kwargs):
@@ -105,7 +109,7 @@ class ApplyDocumentScoringTransformer(ApplyTransformerBase):
             outputRes["score"] = pd.Series(dtype='float64')
             return outputRes
         if self.verbose:
-            tqdm.pandas(desc="pt.apply.doc_score", unit="d")
+            pt.tqdm.pandas(desc="pt.apply.doc_score", unit="d")
             outputRes["score"] = outputRes.progress_apply(fn, axis=1).astype('float64')
         else:
             outputRes["score"] = outputRes.apply(fn, axis=1).astype('float64')
@@ -128,10 +132,9 @@ class ApplyDocumentScoringTransformer(ApplyTransformerBase):
 
         import math
         from .model import split_df
-        from . import tqdm
         num_chunks = math.ceil( len(inputRes) / self.batch_size )
         iterator = split_df(inputRes, num_chunks)
-        iterator = tqdm(iterator, desc="pt.apply", unit='row')
+        iterator = pt.tqdm(iterator, desc="pt.apply", unit='row')
         rtr = pd.concat([self._transform_batchwise(chunk_df) for chunk_df in iterator])
         rtr = add_ranks(rtr)
         return rtr
@@ -146,7 +149,7 @@ class ApplyDocFeatureTransformer(ApplyTransformerBase):
             def _feature_fn(row):
                 return numpy.array([len(row["url"], row["url".count("/")])
             
-            pipe = pt.BatchRetrieve(index) >> pt.apply.doc_features(_feature_fn) >> pt.LTRpipeline(xgBoost())
+            pipe = pt.terrier.Retriever(index) >> pt.apply.doc_features(_feature_fn) >> pt.LTRpipeline(xgBoost())
     """
     def __init__(self, fn,  *args, **kwargs):
         """
@@ -162,7 +165,7 @@ class ApplyDocFeatureTransformer(ApplyTransformerBase):
         fn = self.fn
         outputRes = inputRes.copy()
         if self.verbose:
-            tqdm.pandas(desc="pt.apply.doc_features", unit="d")
+            pt.tqdm.pandas(desc="pt.apply.doc_features", unit="d")
             outputRes["features"] = outputRes.progress_apply(fn, axis=1)
         else:
             outputRes["features"] = outputRes.apply(fn, axis=1)
@@ -178,11 +181,11 @@ class ApplyQueryTransformer(ApplyTransformerBase):
             def _rewriting_fn(row):
                 return row["query"] + " extra words"
             
-            pipe = pt.apply.query(_rewriting_fn) >> pt.BatchRetrieve(index)
+            pipe = pt.apply.query(_rewriting_fn) >> pt.terrier.Retriever(index)
 
         Similarly, a lambda function can also be used::
 
-            pipe = pt.apply.query(lambda row: row["query"] + " extra words") >> pt.BatchRetrieve(index)
+            pipe = pt.apply.query(lambda row: row["query"] + " extra words") >> pt.terrier.Retriever(index)
 
         In the resulting dataframe, the previous query for each row can be found in the query_0 column.
 
@@ -200,13 +203,24 @@ class ApplyQueryTransformer(ApplyTransformerBase):
 
     def transform(self, inputRes):
         from .model import push_queries
-        fn = self.fn
-        outputRes = push_queries(inputRes.copy(), inplace=True, keep_original=True)
-        if self.verbose:
-            tqdm.pandas(desc="pt.apply.query", unit="d")
-            outputRes["query"] = outputRes.progress_apply(fn, axis=1)
+        fn = self.fn        
+        if "query" in inputRes.columns:
+            # we only push if a query already exists
+            outputRes = push_queries(inputRes.copy(), inplace=True, keep_original=True)
         else:
-            outputRes["query"] = outputRes.apply(fn, axis=1)
+            outputRes = inputRes.copy()
+        try:
+            if self.verbose:
+                pt.tqdm.pandas(desc="pt.apply.query", unit="d")
+                outputRes["query"] = outputRes.progress_apply(fn, axis=1)
+            else:
+                outputRes["query"] = outputRes.apply(fn, axis=1)
+        except ValueError as ve:
+            msg = str(ve)
+            if "Columns must be same length as key" in msg:
+                raise TypeError("Could not coerce return from pt.apply.query function into a list of strings. Check your function returns a string.") from ve
+            else:
+                raise ve
         return outputRes
 
 class ApplyGenericTransformer(ApplyTransformerBase):
@@ -225,7 +239,7 @@ class ApplyGenericTransformer(ApplyTransformerBase):
         # this pipeline would remove all but the first two documents from a result set
         lp = ApplyGenericTransformer(lambda res : res[res["rank"] < 2])
 
-        pipe = pt.BatchRetrieve(index) >> lp
+        pipe = pt.terrier.Retriever(index) >> lp
 
     """
 
@@ -248,11 +262,21 @@ class ApplyGenericTransformer(ApplyTransformerBase):
         # batching
         import math, pandas as pd
         from pyterrier.model import split_df
-        from . import tqdm
         num_chunks = math.ceil( len(inputRes) / self.batch_size )
         iterator = split_df(inputRes, num_chunks)
         if self.verbose:
-            iterator = tqdm(iterator, desc="pt.apply", unit='row') 
+            iterator = pt.tqdm(iterator, desc="pt.apply", unit='row') 
         rtr = pd.concat([self.fn(chunk_df) for chunk_df in iterator])
         return rtr
 
+class ApplyIndexer(Indexer):
+    """
+    Allows arbitrary indexer pipelines components to be written as functions.
+    """
+    
+    def __init__(self, fn,  *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.fn = fn
+
+    def index(self, iter_dict):
+        return self.fn(iter_dict)
\ No newline at end of file
