@@ -1,13 +1,17 @@
+import re
 import os
+from types import GeneratorType
+import xml.etree.ElementTree as ET
+import numpy as np
 import pandas as pd
 from contextlib import contextmanager
+import pyterrier as pt
 
 
 def coerce_dataframe(obj):
     if isinstance(obj, pd.DataFrame):
         return obj
-    import types
-    if isinstance(obj, types.GeneratorType):
+    if isinstance(obj, GeneratorType):
         #its a generator, lets assume it generates dataframes
         rtr=[]
         for x in obj:
@@ -41,7 +45,6 @@ def find_files(dir):
     Returns:
         paths(list): A list of the paths to the files
     """
-    import os
     lst = []
     files = []
     for (dirpath, dirnames, filenames) in os.walk(dir, followlinks=True):
@@ -131,7 +134,6 @@ def touch(fname, mode=0o666, dir_fd=None, **kwargs):
     Eqiuvalent to touch command on linux.
     Implementation from https://stackoverflow.com/a/1160227
     """
-    import os
     flags = os.O_CREAT | os.O_APPEND
     with os.fdopen(os.open(fname, flags=flags, mode=mode, dir_fd=dir_fd)) as f:
         os.utime(f.fileno() if os.utime in os.supports_fd else fname,
@@ -179,7 +181,6 @@ def read_results(filename, format="trec", topics=None, dataset=None, **kwargs):
     if dataset is not None:
         assert topics is None, "Cannot provide both dataset and topics"
         if isinstance(dataset, str):
-            import pyterrier as pt
             dataset = pt.get_dataset(dataset)
         topics = dataset.get_topics()
     if topics is not None:
@@ -195,8 +196,6 @@ def _read_results_letor(filename, labels=False):
             # my $label = shift @parts;
             # my %hash = map {split /:/, $_} @parts;
             # return ($label, $comment, %hash);
-        import re
-        import numpy as np
         line, comment = l.split("#")
         line = line.strip()
         parts = re.split(r'\s+|:', line)
@@ -228,12 +227,8 @@ def _read_results_letor(filename, labels=False):
 
 def _read_results_trec(filename):
     results = []
-    df = pd.read_csv(filename, sep=r'\s+', names=["qid", "iter", "docno", "rank", "score", "name"])
+    df = pd.read_csv(filename, sep=r'\s+', names=["qid", "iter", "docno", "rank", "score", "name"], dtype={'qid': str, 'docno': str, 'rank': int, 'score': float}) 
     df = df.drop(columns="iter")
-    df["qid"] = df["qid"].astype(str)
-    df["docno"] = df["docno"].astype(str)
-    df["rank"] = df["rank"].astype(int)
-    df["score"] = df["score"].astype(float)
     return df
 
 def write_results(res, filename, format="trec", append=False, **kwargs):
@@ -294,20 +289,19 @@ def read_topics(filename, format="trec", **kwargs):
 
     Supported Formats:
         * "trec" -- an SGML-formatted TREC topics file. Delimited by TOP tags, each having NUM and TITLE tags; DESC and NARR tags are skipped by default. Control using whitelist and blacklist kwargs
-        * "trecxml" -- a more modern XML formatted topics file. Delimited by topic tags, each having nunber tags. query, question and narrative tags are parsed by default. Control using tags kwarg.
-        * "singeline" -- one query per line, preceeded by a space or colon. Tokenised by default, use tokenise=False kwargs to prevent tokenisation.
+        * "trecxml" -- a more modern XML formatted topics file. Delimited by topic tags, each having number tags. query, question and narrative tags are parsed by default. Control using tags kwarg.
+        * "singleline" -- one query per line, preceeded by a space or colon. Tokenised by default, use tokenise=False kwargs to prevent tokenisation.
     """
     if format is None:
         format = "trec"
     if not format in SUPPORTED_TOPICS_FORMATS:
-        raise ValueError("Format %s not known, supported types are %s" % (format, str(SUPPORTED_RESULTS_FORMATS.keys())))
+        raise ValueError("Format %s not known, supported types are %s" % (format, str(SUPPORTED_TOPICS_FORMATS.keys())))
     return SUPPORTED_TOPICS_FORMATS[format](filename, **kwargs)
 
+@pt.java.required
 def _read_topics_trec(file_path, doc_tag="TOP", id_tag="NUM", whitelist=["TITLE"], blacklist=["DESC","NARR"]):
-    from jnius import autoclass
-    from . import check_version
-    assert check_version("5.3")
-    trecquerysource = autoclass('org.terrier.applications.batchquerying.TRECQuery')
+    assert pt.terrier.check_version("5.3")
+    trecquerysource = pt.java.autoclass('org.terrier.applications.batchquerying.TRECQuery')
     tqs = trecquerysource(
         [file_path], doc_tag, id_tag, whitelist, blacklist,
         # help jnius select the correct constructor
@@ -320,6 +314,7 @@ def _read_topics_trec(file_path, doc_tag="TOP", id_tag="NUM", whitelist=["TITLE"
     topics_dt = pd.DataFrame(topics_lst,columns=['qid','query'])
     return topics_dt
 
+@pt.java.required
 def _read_topics_trecxml(filename, tags=["query", "question", "narrative"], tokenise=True):
     """
     Parse a file containing topics in TREC-like XML format
@@ -330,16 +325,16 @@ def _read_topics_trecxml(filename, tags=["query", "question", "narrative"], toke
     Returns:
         pandas.Dataframe with columns=['qid','query']
     """
-    import xml.etree.ElementTree as ET
-    import pandas as pd
     tags=set(tags)
     topics=[]
     tree = ET.parse(filename)
     root = tree.getroot()
-    from jnius import autoclass
-    tokeniser = autoclass("org.terrier.indexing.tokenisation.Tokeniser").getTokeniser()
+    tokeniser = pt.java.autoclass("org.terrier.indexing.tokenisation.Tokeniser").getTokeniser()
     for child in root.iter('topic'):
-        qid = child.attrib["number"]
+        try:
+            qid = child.attrib["number"]
+        except KeyError:
+            qid = child.find("number").text
         query = ""
         for tag in child:
             if tag.tag in tags:
@@ -347,9 +342,10 @@ def _read_topics_trecxml(filename, tags=["query", "question", "narrative"], toke
                 if tokenise:
                     query_text = " ".join(tokeniser.getTokens(query_text))
                 query += " " + query_text
-        topics.append((str(qid), query))
+        topics.append((str(qid), query.strip()))
     return pd.DataFrame(topics, columns=["qid", "query"])
 
+@pt.java.required
 def _read_topics_singleline(filepath, tokenise=True):
     """
     Parse a file containing topics, one per line. This function uses Terrier, so supports reading direct from URLs.
@@ -363,10 +359,8 @@ def _read_topics_singleline(filepath, tokenise=True):
         pandas.Dataframe with columns=['qid','query']
     """
     rows = []
-    from jnius import autoclass
-    from . import check_version
-    assert check_version("5.3")
-    slqIter = autoclass("org.terrier.applications.batchquerying.SingleLineTRECQuery")(filepath, tokenise)
+    assert pt.terrier.check_version("5.3")
+    slqIter = pt.java.autoclass("org.terrier.applications.batchquerying.SingleLineTRECQuery")(filepath, tokenise)
     for q in slqIter:
         rows.append([slqIter.getQueryId(), q])
     return pd.DataFrame(rows, columns=["qid", "query"])
@@ -400,3 +394,17 @@ SUPPORTED_RESULTS_FORMATS = {
     "letor" : (_read_results_letor, _write_results_letor),
     "minimal" : (None, _write_results_minimal)
 }
+
+
+def pyterrier_home() -> str:
+    """
+    Returns pyterrier's home directory. By default this is ~/.pyterrier, but it can also be set with the PYTERRIER_HOME
+    env variable.
+    """
+    if "PYTERRIER_HOME" in os.environ:
+        home = os.environ["PYTERRIER_HOME"]
+    else:
+        home = os.path.expanduser('~/.pyterrier')
+    if not os.path.exists(home):
+        os.makedirs(home)
+    return home
